from sqlalchemy import Column, ForeignKey, func
from sqlalchemy import Integer, String, DateTime, Boolean, UniqueConstraint
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship, backref

Base = declarative_base()

class Processors(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "processors"
    id = Column(Integer, primary_key=True)
    name = Column(String(30))
    uri = Column(String())
    rel_path = Column(String())
    added = Column(DateTime, default = func.now())

    # joins
    #uris = relationship("URIs", secondary= "processoruris") #back_populates = "processor")
    pmd = relationship("ProcessorMD")

    # contstaints
    UniqueConstraint("processor_relpath", "rel_path", name='prel_path_unique')
    UniqueConstraint("processor_uri", "uri_path", name='puri_unique')

    def __repr__(self):
        return f"Processor(id={self.id!r}, name={self.name!r})"

class ProcessorMD(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "processormd"
    processorid = Column(Integer, ForeignKey("processors.id"), primary_key = True)
    key = Column(String(255), primary_key = True)
    added = Column(DateTime, default = func.now())
    value = Column(String())

    def __repr__(self):
        return f"Processor MD(id={self.id!r}, key={self.key!r}, value={self.value!r})"

class Domains(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "domains"
    id = Column(Integer, primary_key=True)
    domain = Column(String())
    domain_type = Column(String(15))

    # joins
    dmd = relationship("DomainMD")
    domains = relationship("URIs")

    # contstaints
    UniqueConstraint("domain_domain", "domain", name='ddomain_unique')

    def __repr__(self):
        return f"Domain(id={self.id!r}, name={self.domain!r})"

class DomainMD(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "domainmd"
    domainid = Column(Integer, ForeignKey("domains.id"), primary_key = True)
    key = Column(String(255), primary_key = True)
    added = Column(DateTime, default = func.now())
    modified = Column(DateTime)
    value = Column(String())

    def __repr__(self):
        return f"Domain MD(id={self.id!r}, key={self.key!r}, value={self.value!r})"

class URIs(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "uris"
    id = Column(Integer, primary_key=True)
    uriid_parent = Column(Integer, default=None) # , ForeignKey("uris.id")

    uri = Column(String())

    uri_type = Column(String(15))
    domainid = Column(Integer, ForeignKey("domains.id"))
    root = Column(Boolean, default = False)
    added = Column(DateTime, default = func.now())
    deleted = Column(DateTime)

    # joins
    #processors = relationship("Processors", secondary= "processoruris") # back_populates = "uri")
    umd = relationship("URIMD")
    versions = relationship("Versions")

    # contstaints
    UniqueConstraint("uriid_parent", "uri", name='uri_unique')

    def __repr__(self):
        return f"URI(id={self.id!r}, name={self.uri!r})"

class URIMD(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "urimd"
    uriid = Column(Integer, ForeignKey("uris.id"), primary_key = True)
    key = Column(String(255), primary_key = True)
    added = Column(DateTime, default = func.now())
    modified = Column(DateTime)
    deleted = Column(DateTime)
    value = Column(String())

    def __repr__(self):
        return f"URI MD(uriid={self.uriid!r}, key={self.key!r}, value={self.value!r})"

class ProcessorURIs(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "processoruris"
    #id = Column(Integer, primary_key=True)
    processorid = Column(Integer, ForeignKey("processors.id"), primary_key = True)
    uriid = Column(Integer, ForeignKey("uris.id"), primary_key = True)
    added = Column(DateTime, default = func.now())
    deleted = Column(DateTime)

    # joins
    uri = relationship("URIs", backref = backref("processors_assoc"))
    processor = relationship("Processors", backref = backref("uris_assoc"))

class Versions(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "versions"
    id = Column(Integer, primary_key=True)
    uriid = Column(Integer, ForeignKey("uris.id"))
    added = Column(DateTime, default = func.now())
    modified = Column(DateTime)
    size = Column(Integer)
    checksum = Column(String(255))

    # join
    vmd = relationship("VersionMD")

    def __repr__(self):
        return f"URI(versionid={self.id!r}, uriid={self.uriid!r}, size={self.size}, modified={self.modified})"

class VersionMD(Base):
    __table_args__ = {'extend_existing': True}
    __tablename__ = "versionmd"
    id = Column(Integer, primary_key=True)
    versionid = Column(Integer, ForeignKey("versions.id"))
    added = Column(DateTime, default = func.now())
    modified = Column(DateTime)
    deleted = Column(DateTime)
    key = Column(String(255))
    value = Column(String())

    def __repr__(self):
        return f"Version MD(uriid={self.versionid!r}, key={self.key!r}, value={self.value!r})"



"""
Lost cat will scan and process a range of files
"""
import logging
import multiprocessing as mp
import shelve
import time

from datetime import datetime
from queue import Empty
from sqlite3 import IntegrityError
from sqlalchemy import func

from lost_cat.database.db_utils import DBEngine
from lost_cat.database.schema import URIMD, Domains, DomainMD, ProcessorMD, \
                ProcessorURIs, Processors, URIs, VersionMD, Versions
from lost_cat.utils.module_utils import load_module, load_modulefile
from lost_cat.utils.path_utils import SourceNotValid
from lost_cat.utils.tag_anon import TagAnon

logger = logging.getLogger(__name__)

# <TODO: move this out exception library>
class FeatureNotImplemented(Exception):
    """used for the feature not implemented"""
    def __init__(self, label: str, feature: str, message: str) -> None:
        super().__init__()
        self.label = label
        self.feature = feature
        self.message = message

class ProcessorCannotbeLoaded(Exception):
    """A simple exception to raise already exist error"""

class SourceAlreadyExists(Exception):
    """A simple exception to raise already exist error"""

class ParserAlreadyExists(Exception):
    """A simple exception to raise already exist error"""

class ProcessorAlreadyExists(Exception):
    """A simple exception to raise already exist error"""

class NoHandlersFoundForType(Exception):
    """A simple exception to raise already exist error"""
    def __init__(self, label: str, message: str) -> None:
        super().__init__()
        self.label = label
        self.message = message

class ClassFailedToLoad(Exception):
    """A simple exception to raise already exist error"""
    def __init__(self, label: str, message: str) -> None:
        super().__init__()
        self.label = label
        self.message = message

class ParserFailedToLoad(Exception):
    """A simple exception to raise already exist error"""
    def __init__(self, label: str, base_class: str, message: str) -> None:
        super().__init__()
        self.label = label
        self.base_class = base_class
        self.message = message

class LostCat():
    """
    The Lost Cat Main class

    - Lost cat will accept a range of sources (OS Paths)
    - Scan the folders and files
    - Create an artifact list
    - Create a catalog:
        - Grouped by key tags
        - metadata extracted from files as needed
        - file / folder path metadata included
        - it will also scan and index archive files (zip only atm)
    - provide a set of tools to move, relabel, and so on to help organize
    """
    name = "LostCat"
    version = "0.1.0"

    def __init__(self,
            options: dict = None,
            paths: dict = None) -> None:
        """Initialize the core elements

        Parameters
        ----------
        opttion : dict      A dict that sets the search options for the syste,
            {
                "threads": {

                },
                "<scanner - class name>": {
                    scanner options - see scanner classes for supported options
                },
                "<parsers - class name>": {
                    parser options - see parser classes for support options
                },
                "<converters - class name>" :{
                    converter options - see converter classes for support options
                }
            }

        paths: dict  A dict of paths to use for the components
            {
                "artifacts":    path to artifacts shelve file
                "databasae":    path to the sqlite db
            }

        """
        # a labelled dic of sources,
        # sources are parsed to an object
        self._sources = dict()          # a dict via labels, of uris to scan
        self._processors = dict()         # a dict of the available processors to use
                                        # the processor will only
        self._urihandlers = dict()      # a dictionary of the uri types and the list of
                                        # processor labels to use
        self._fltrhandlers = dict()  # a dictionary of the uri subtypes and the list of
                                        # processor labels to use

        self._parsers = dict()          # a dict of parser, grouped by ext / type
        self._parse_ext = dict()        # a dict of extension used to select the
                                        # correct parser to use
        self._features = ["scanner", "parser", "converter"]     # a set of features to make avaiable
        self._anonobj = None            # the anonymizer class, will look at tags and
                                        # them, using a local cache for consistency

        # caches for updates...
        self._domains = {}              # K: domain          | V: id
        self._uris = {}                 # K: uri             | V: id
        self._urimd = {}                # K: uriid, key      | V: id, value
        self._versions = {}             # K: uriid, modified | V: row
        self._vlatest = {}              # KL uriid           | v: row
        self._versionmd = {}            # K: versionid, key  | V: id, value

        # tags might be better to be set at the scanner, parser level
        self._tags = dict()             # the provided tags

        # set the objects
        if options:
            self._options = options
        else:
            # default to create a phrase profile for the filename
            self._options = {
                "threads": {
                    "count": 3,
                    "stop": True, # tells the system to stop when the in queue is empty
                    "timeout": 2
                }
            }

        # shelve root is the base oath for the shelve file
        # this program will create a shelve for each function run...
        # artifacts - augmented with meatadata and file information
        self._paths = paths
        self._artifacts = None
        self._db = None

        if "artifacts" in paths:
            # a local store for the disovered artifacts
            logger.info("Shelve file: %s", self._paths.get("artifacts", "<missing>"))
            self._artifacts = shelve.open(self._paths.get("artifacts", "<missing>"))

        else:
            logger.info("Database: %s", self._paths.get("database", "<missing>"))
            self._db = DBEngine(connString=self._paths.get("database", "lost-cat.db"))
            self.load_processors()

        # a place to store the processed artifacts, organized
        # by the grouping, and with metadata...
        self._catalog = dict()

    def close(self) -> None:
        """Will save the shelve to the OS"""
        if self._artifacts:
            logger.debug("Closing shelve %s", len(self._artifacts))
            self._artifacts.close()

    def set_tags(self, tags: dict) -> None:
        """Accepts a dictionary of tags expecets:
            alias: dict of k,v pairs, k is field to convert v is the field name
            groups: list of tags to use for grouping
            metadata: list of tags to metadata
            anonimizer: list of tags to run through the anonimizer
            <class name>: { alias, group, metadata}
        """
        logger.debug("Added tags: %s", tags)
        self._tags = tags

    def set_anonimizer(self, anonimizer: TagAnon) -> None:
        """This will set the anonimizer for the class"""
        self._anonobj = anonimizer

    def load_processors(self):
        """Will load the core information from the underlying datastores"""
        if not self._db:
            return

        _db_sess = self._db.session()

        self._processors = {}

        # load processort metadata if any
        pmd_results = _db_sess.query(Processors.id,
                    Processors.uri,
                    Processors.name,
                    ProcessorMD.key,
                    ProcessorMD.value).join(ProcessorMD).all()
        logger.info("Processors MD: %s", pmd_results)

        _pmd = {}
        for r in pmd_results:
            logger.info("Row: %s", r)
            if not r.name:
                continue
            _pname = r.name
            if _pname not in _pmd:
                _pmd[_pname] = {}
            _pmd[_pname][r.key] = r.value

        # load in the tables from the source system...
        for r in _db_sess.query(Processors).all():
            logger.info("Proc Row: %s", r)
            try:
                if not r.name:
                    continue

                _pname = r.name
                if _pname not in self._processors:
                    _pname, _uri, base_class, obj = self.resolve_class(
                                module_path=r.uri)
                    if obj:
                        obj.processorid = r.id

                    self._processors[_pname] = {
                        "id": r.id,
                        "uri": _uri,
                        "class": base_class,
                        "obj": obj,
                        "metadata": _pmd.get(_pname),
                    }

                    _tags = None
                    if not (_tags := self._tags.get("pname")):
                        # load the default _tags
                        if self._tags:
                            _tags ={
                                "alias": self._tags.get("alias", {}),
                                "groups": self._tags.get("groups", {}),
                                "metadata": self._tags.get("metadata", {}),
                                "anonimizer": self._tags.get("anonimizer", {}),
                            }

                    if _tags:
                        self._processors[_pname]["tags"] = _tags

                    self.add_typehandler(obj, _pname)

            except Exception as ex:
                logger.error("Failed to load processor %s",  ex)

        logger.info(self._processors)

    def add_processor(self, label: str = None,
            base_class: object = None,
            module_path: str = None,
            settings: dict = None,
            overwrite: bool = False):
        """will add a processor to the the system, and perform simple
        setting and config loads."""

        _pname, _uri, base_class, obj = self.resolve_class(
                        base_class=base_class,
                        module_path=module_path,
                        settings=settings,
                        overwrite=overwrite)
        _pmd = {
            _pname: {}
        }

        if _pname in self._processors:
            logger.info("Updating class: %s %s", _pname, _uri)
            self._processors[_pname]["base_class"] = base_class
            self._processors[_pname]["obj"] = obj

        else:
            # add the processor to the database
            _db_sess = self._db.session()
            _proc = _db_sess.query(Processors).where(Processors.name == _pname).one_or_none()

            _procmds = []
            if _proc:
                # get the metadata
                for _procmd in list(_db_sess.query(ProcessorMD).where(ProcessorMD.processorid == _proc.id).all()):
                    logger.debug("Proc MD: %s", _procmd)

            else:
                logger.info("Adding class: %s %s", _pname, _uri)

                try:
                    _proc = Processors(name=_pname, uri=_uri)
                    _db_sess.add(_proc)
                    _db_sess.flush()

                    logger.info(_proc)

                    # Add the metadata elements if there are any
                    for _mdk, _mdv in obj.avail_config().get("metadata",{}).items():
                        _proc_md = ProcessorMD(processorid =_proc.id, key = _mdk, value = _mdv)
                        _pmd[_pname][_mdk] = _mdv
                        _procmds.append(_proc_md)

                    _db_sess.add_all(_procmds)
                    _db_sess.flush()
                    _db_sess.commit()

                    if obj:
                        obj.processorid = _proc.id

                except Exception as ex:
                    logger.error("Error adding processor %s %s %s", base_class, module_path, ex)
                    _db_sess.rollback()

                    for row in _db_sess.query(Processors).where(Processors.uri == _uri):
                        logger.info(row)
                        _proc = row

                    logger.info(_proc)

            self._processors[_pname] = {
                        "id": _proc.id,
                        "uri": _proc.uri,
                        "class": base_class,
                        "obj": obj,
                        "metadata": _pmd.get(_pname),
                    }

            _tags = None
            if not (_tags := self._tags.get("pname")):
                # load the default _tags
                if self._tags:
                    _tags ={
                        "alias": self._tags.get("alias", {}),
                        "groups": self._tags.get("groups", {}),
                        "metadata": self._tags.get("metadata", {}),
                        "anonimizer": self._tags.get("anonimizer", {}),
                    }

            if _tags:
                self._processors[_pname]["tags"] = _tags

        self.add_typehandler(obj, _pname)
        #self.add_filterhandler(obj, _pname)

    def resolve_class(self, base_class: object = None,
                        module_path: str = None,
                        rel_path: str = None,
                        settings: dict = None,
                        overwrite: bool = False):
        """ """

        _cls_name = module_path.split('.')[-1] if module_path else None
        logger.info("Name: %s", _cls_name)

        try:
            if module_path:
                logger.info("Loading path: %s", module_path)
                obj = None
                if rel_path:
                    _objs = load_modulefile(folder=module_path, file=rel_path)

                else:
                    _objs = load_module(module = module_path)

                logger.info("OBJS: %s", _objs)
                for _o, _v in _objs.items():
                    logger.info("\t%s => %s", _o, _v)

                _base_class = _objs.get(_cls_name)
                logger.info("Class: %s", _base_class)

                if _base_class:
                    base_class = _base_class
            elif base_class:
                logger.info("Loading obj: %s", base_class)

        except Exception as ex:
            raise ClassFailedToLoad(label=module_path,
                    message="class could not be loaded MODULE") from ex

        if not base_class:
            raise ClassFailedToLoad(label=module_path,
                    message="class could not be loaded MODULE")

        # load the class and extract the
        try:
            obj = base_class(settings=settings)
            _pname = obj.name
            _uri = f"{obj.__module__}.{obj.__class__.__name__}"

        except ImportError as ex:
            raise ClassFailedToLoad(label=base_class,
                    message="class could not be loaded BASE_CLASS") from ex

        # save the processor to the database
        if _pname in self._processors and overwrite is False:
            logger.debug("Class is already loaded %s %s", _pname, _uri)
            return _pname, _uri, base_class, obj

        return _pname, _uri, base_class, obj

    def add_typehandler(self, obj, _pname):
        """ now to set the uritypes it can work with... """
        for _uritype in obj.avail_config().get("uritypes",[]):
            if _uritype not in self._urihandlers:
                self._urihandlers[_uritype] = []
            if _pname not in self._urihandlers[_uritype]:
                self._urihandlers[_uritype].append(_pname)

    def get_sourcequery(self, filters: dict, urimd: list = None, versionmd: list = None, domains: list = None) -> list:
        """Will return a sql string used to query the underlying db
        For example
            SELECT
                u.*,
                pumd.*
            FROM uris u

            -- for the metadata
            INNER JOIN (SELECT
                ium.uriid AS uriid,
                MAX(CASE WHEN ium.key = 'ext' THEN ium.value END) AS ext,
                MAX(CASE WHEN ium.key = 'name' THEN ium.value END) AS name,
                MAX(CASE WHEN ium.key = 'zipfile' THEN ium.value END) AS zipfile
            FROM
                urimd ium
            GROUP BY ium.uriid) AS pumd
                ON u.id = pumd.uriid

            -- for the filter
            INNER JOIN urimd um
                ON u.id = um.uriid AND um.key = 'ext' AND um.value IN ('.dcm')
            WHERE u.domainin in ()
        """

        sel_q = """SELECT u.id AS uriid,
            u.uri_type AS uritype,
            u.uri AS uri"""

        _sel_umd = ""
        _j_umd = ""
        _sel_vmd = ""
        _j_vmd = ""

        if urimd:
            _md_qry = []
            for _md in urimd:
                _md_qry.append(f"MAX(CASE WHEN ium.key = '{_md}' THEN ium.value END) AS {_md}")

            _sel_umd = ", pumd.* "
            _pvt_umd = ", " + ", ".join(_md_qry)
            _umd_qry = f""" INNER JOIN (SELECT ium.uriid AS uriid{_pvt_umd} FROM urimd AS ium GROUP BY ium.uriid) AS pumd ON u.id = pumd.uriid"""

        if versionmd:
            _vmd_qry = []
            for _vmd in versionmd:
                _vmd_qry.append(f"MAX(CASE WHEN vm.key = '{_vmd}' then vm.value END) AS {_vmd}")

            _sel_vmd = _sel_vmd = ", " + ", ".join(_vmd_qry)
            _vmd_qry = " INNER JOIN Versions v ON u.id = v.uriid INNER JOIN VersionsMD vm on v.id = vm.versionid "

        from_q = {
            "URIs": """{q}{sumd}{svmd} FROM URIS u {umd}
                    WHERE u.{field} == {select}""",
            "URIMD":"""{q}{sumd}{svmd} FROM URIS u {umd}
                    INNER JOIN URIMD um ON u.id = um.uriid
                    WHERE um.{field} == \"{select}\"
                    AND um.value IN ({filter})""",
            "VersionsMD": """{q}{sumd}{svmd} FROM URIS u {umd}{vmd}
                    INNER JOIN Versions v ON u.id = v.uriid
                    INNER JOIN VersionsMD vm on v.id = vm.versionid
                    WHERE vm.{field} == \"{select}\"
                    AND vm.value IN ({filter})"""
        }

        _queries = []

        for _fltrtype in filters:
            # table, field, select, filter
            _tbl = _fltrtype.get("table")
            _fld = _fltrtype.get("field")
            _sel = _fltrtype.get("select")
            logger.debug("Filter: %s", _fltrtype)

            _vals = {
                "q": sel_q,
                "field": _fld,
                "select": _sel,
                "sumd": _sel_umd if urimd else "",
                "umd": _umd_qry if urimd else "",
                "svmd": _sel_vmd if urimd else "",
                "vmd": _vmd_qry if versionmd else ""
            }

            #
            if "filter" not in _fltrtype:
                _sql = from_q.get(_tbl,"") \
                        .format(**_vals)
                _queries.append(_sql)
            else:
                _vals["filter"] = "\"{}\"".format(",\"".join(_fltrtype.get("filter",[])))
                _sql = from_q.get(_tbl,"") \
                        .format(**_vals)
                _queries.append(_sql)

            logger.debug("SQ SQL: %s", _sql)

        return _queries

    def load_db_sources(self):
        """will read the sources from the db"""
        if not self._db:
            return

        # load in the tables from the source system...
        """type
                processors []
                uris
                    uri
        """
        self._sources = {}
        sql = """SELECT
                    uris.id AS uriid,
                    uris.uri AS uri,
                    uris.uri_type AS uritype,
                    uris.root AS uriroot,
                    uris.added AS uriadded,
                    uris.deleted AS urideleted,
                    processors.id AS processorid,
                    processors.name AS processorname,
                    processors.uri AS processoruri
                FROM processoruris
                JOIN uris
                    ON uris.id = processoruris.uriid
                JOIN processors
                    ON processors.id = processoruris.processorid
                WHERE uris.root = 1
                    AND uris.deleted IS NULL
                ORDER BY uris.uri_type"""

        for row in self._db.sql(sql=sql):
            # add to the sources dict'
            logger.debug(row)
            _uri_type = row.get("uritype")
            _processor = row.get("processorname")
            _uri = row.get("uri")

            if _uri_type not in self._sources:
                self._sources[_uri_type] = {
                    "processors": [],
                    "uris": {}
                }

            # add the handler to the type
            if _processor not in self._sources[_uri_type]["processors"]:
                self._sources[_uri_type]["processors"].append(_processor)

            # add the uri
            if _uri not in self._sources[_uri_type]["uris"]:
                self._sources[_uri_type]["uris"][_uri] = row.copy()

        #logger.debug(self._sources)

    def add_source(self, processor: str, uri: str,
                isroot: bool = False, overwrite: bool = False) -> dict:
        """It parses the provided source path and
        add to the source list.Build path returns a dict of the uri:
            "type": str
            "uri: str
        """
        logger.debug("Processor: %s URI: %s ", processor, uri)
        if _processor := self._processors.get(processor):
            try:
                _processor["obj"] = _processor.get("class")()
                _uri_obj = _processor.get("obj").build_path(uri=uri)

            except Exception as ex:
                raise ClassFailedToLoad(label=uri,
                        message="Class provided could not be loaded",
                    ) from ex
        else:
            raise ClassFailedToLoad(label=uri,
                    message="Class provided could not be loaded",
                )

        if not _uri_obj:
            raise SourceNotValid(uri=uri, message="missing processor record!")

        logger.debug("Processor: %s", _uri_obj)
        _uritype = _uri_obj.get("type","<>")

        if uri in self._sources.get(_uritype,{}).get("uris",{}) and overwrite is False:
            return self._sources.get(_uritype,{}).get("uris",{}).get(uri,{})

        # add the uri to the sources...
        _procid = _processor.get("id")
        _db_sess = self._db.session()

        # add the domain if needed...
        try:
            _domain = _db_sess.query(Domains).filter(Domains.domain == _uri_obj.get("domain")).one_or_none()
            if not _domain:
                _domain = Domains(domain=_uri_obj.get("domain"), domain_type=_uri_obj.get("domain_type"))
                _db_sess.add(_domain)
                _db_sess.commit()

            _domid = _domain.id

        except Exception as ex:
            logger.error("ERROR adding domain %s %s", _uri_obj, ex)
            # fetch the URI record...
            _db_sess.rollback()

        # add the uri
        try:
            _uri = _db_sess.query(URIs).filter(URIs.uri == uri).one_or_none()
            if not _uri:
                _uri = URIs(uri=uri,
                    uri_type =_uri_obj.get("type"),
                    root = isroot,
                    domainid = _domid,
                    uriid_parent = None)

                _db_sess.add(_uri)
                _uriid = _uri.id
                _db_sess.commit()

        except Exception as ex:
            logger.error("ERROR adding uri %s %s", uri, ex)
            # fetch the URI record...
            _db_sess.rollback()
            for row in list(_db_sess.query(URIs).where(URIs.uri == uri).all()):
                logger.debug(row)
                _uri = row

            logger.debug(_uri)
            _uriid = _uri.id

        try:
            _procuri = _db_sess.query(ProcessorURIs).where(ProcessorURIs.uriid == _uri.id, ProcessorURIs.processorid == _procid).one_or_none()
            if not _procuri:
                _procuri = ProcessorURIs(processorid = _procid, uriid = _uri.id)
                _db_sess.add(_procuri)
                _db_sess.commit()

        except Exception as ex:
            logger.error("ERROR adding processor uri %s %s %s", _procid, uri, ex)
            _db_sess.rollback()

        _uri_obj["uriid"] = _uri.id
        _uri_obj["processorid"] = _procid
        _uri_obj["processorname"] = processor

        # add to teh source dictionary for root items
        if isroot is True:
            if _uritype not in self._sources:
                self._sources[_uritype] = {
                    "processors": [_processor.get("name")],
                    "uris": {}
                }

            self._sources[_uritype]["uris"][uri] = _uri_obj

        return _uri_obj

    def save_queue(self, _out_queue):
        """ Save the data from the queue,
        structured by the following:
            processorid
            domain / md
            uri / md
            version / md"""

        _stats = {
            "loaded": 0,
            "added": 0,
            "updates": 0,
            "deleted": 0
        }

        _db_sess = self._db.session()
        while _out_queue:
            try:
            # set a timeout, and handle the semiphore case too
                o_item = _out_queue.get(timeout=10) if _out_queue else None
                # URIs
                #   -> Metadata
                #   -> Versions
                #       -> metadata
                if o_item == "DONE":
                    break

                logger.debug("Out: %s", o_item)
                # update the database...
                # Domains
                # K: domain          | V: id
                _domain = _db_sess.query(Domains).filter(Domains.domain == o_item.get("domain")).one_or_none()
                if not _domain:
                    _domain = Domains(
                        domain = o_item.get("domain")
                    )
                    _db_sess.add(_domain)
                    _db_sess.flush()

                _domainid = _domain.id

                # domain metadata...
                _domainmds = _db_sess.query(DomainMD).filter(DomainMD.domainid == _domainid).all()
                _domainmd_keys = {}
                for _domainmd in _domainmds:
                    _domainmd_keys[_domainmd.key] = _domainmd.value

                # update the metadata or add if mssing
                _mdadd = []
                for _mdk, _mdv in o_item.get("domainmd",{}).items():
                    _mdval = _domainmd_keys.get(_mdk)
                    if _mdval:
                        if not _mdv:
                            continue

                        if not _mdval or _mdv != _mdval:
                            # perform an update...
                            _db_sess.query(DomainMD) \
                                    .filter_by(domainid=_domainid, key=_mdk) \
                                    .update({
                                        "value": _mdv if _mdv else "<missing>",
                                        "modified": datetime.now()
                                    })

                            _stats["updates"] += 1
                    else:
                        # insert new value
                        if not _mdv:
                            continue

                        _mdrec = DomainMD(
                            domainid = _domainid,
                            key = _mdk if _mdk else "<missing>",
                            value = _mdv,
                            modified = datetime.now()
                        )
                        _mdadd.append(_mdrec)
                        _stats["added"] += 1

                _db_sess.add_all(_mdadd)
                _db_sess.flush()

                # get the uri
                # check for the parent id...
                # if a file in a zip file get the id of the zipfile...
                # else used the base_id

                # have the source uriid that generated the file
                _uriid_source = o_item.get("uriid_source", 0)

                if _uri_zipfile := o_item.get("metadata",{}).get("zipfile"):
                    # the url for the zip file..
                    if _uri_zf := _db_sess.query(URIs).filter(URIs.uri == _uri_zipfile).one_or_none():
                        _uriid_source = _uri_zf.id
                    else:
                        logger.warning("Unable to fidn zipfile id %s", _uri_zipfile)

                # handle zip files as well
                _uri = _db_sess.query(URIs).filter(
                        URIs.uri == o_item.get("uri"),
                        URIs.uriid_parent == _uriid_source).one_or_none()

                if not _uri:
                    logger.debug("ADD: %s", o_item.get("uri"))
                    _uri = URIs(
                            domainid = _domain.id,
                            uri = o_item.get("uri"),
                            uri_type = o_item.get("uri_type"),
                            uriid_parent = _uriid_source,
                            root=False)
                    _db_sess.add(_uri)
                    _db_sess.flush()

                    _stats["added"] += 1

                _uriid = _uri.id

                # add the relationship to the processor id if included...
                if _processorid := o_item.get("processorid"):
                    if not (_procuri := _db_sess.query(ProcessorURIs).filter(
                                    ProcessorURIs.uriid == _uriid,
                                    ProcessorURIs.processorid == _processorid
                            ).one_or_none()):
                        _procuri = ProcessorURIs(
                                processorid = _processorid,
                                uriid = _uriid,
                                added = datetime.now()
                        )
                        _db_sess.add(_procuri)
                        _db_sess.flush()

                        _stats["added"] += 1

                # get the uri metadata
                _urimds = _db_sess.query(URIMD).filter(URIMD.uriid == _uriid).all()
                _urimd_keys = {}
                for _urimd in _urimds:
                    _urimd_keys[_urimd.key] = _urimd.value

                # update the metadata
                # URIMD
                # K: uriid, key      | V: value
                _mdadd = []
                for _mdk, _mdv in o_item.get("metadata",{}).items():
                    _mdval = _urimd_keys.get(_mdk)
                    if _mdval:
                        if not _mdv:
                            continue

                        if not _mdval or _mdv != _mdval:
                            # perform an update...
                            _db_sess.query(URIMD) \
                                    .filter_by(uriid=_uriid, key=_mdk) \
                                    .update({
                                        "value": _mdv if _mdv else "<missing>",
                                        "modified": datetime.now()
                                    })

                            _stats["updates"] += 1
                    else:
                        # insert new value
                        if not _mdv:
                            continue

                        _mdrec = URIMD(
                            uriid = _uriid,
                            key = _mdk if _mdk else "<missing>",
                            value = _mdv,
                            modified = datetime.now()
                        )
                        _mdadd.append(_mdrec)
                        _stats["added"] += 1

                _db_sess.add_all(_mdadd)
                _db_sess.flush()

                # versions
                if o_item.get("versions", {}).get('__latest__'):
                    _subquery = _db_sess.query(
                        Versions,
                        func.rank().over(
                            order_by=Versions.modified.desc(),
                            partition_by=Versions.uriid
                        ).label('rank')
                    ).filter(Versions.uriid == _uri.id
                    ).order_by(Versions.uriid).subquery()

                    _version = _db_sess.query(_subquery).filter(
                            _subquery.c.rank == 1).one_or_none()

                else:
                    _version = _db_sess.query(Versions).filter(
                            Versions.modified == o_item.get("versions", {}).get('modified'),
                            Versions.uriid == _uri.id).one_or_none()

                    # check for changes
                    if _version and (_version.size != o_item.get("versions", {}).get('size') \
                        or _version.modified != o_item.get("versions", {}).get('modified')):

                        logger.info("updated File...%s", o_item.get("uri"))
                        logger.info("\tOItem: %s", o_item.get("versions", {}))
                        logger.debug("\tOrig: Size: %s\n\t      Mod:  %s", _version.size, _version.modified)
                        logger.debug("\tNew:  Size: %s\n\t      Mod:  %s", o_item.get("versions", {}).get('size'), o_item.get("versions", {}).get('modified'))
                        _version = None

                if not _version:
                    _vmod = o_item.get("versions", {}).get('modified')
                    _version = Versions(
                            uriid = _uriid,
                            modified = o_item.get("versions", {}).get('modified'),
                            size = o_item.get("versions", {}).get('size'))

                    _db_sess.add(_version)
                    _db_sess.flush()

                _vid = _version.id

                # version metadata
                _versionmds = _db_sess.query(VersionMD).filter(VersionMD.versionid == _version.id).all()
                _vmd_keys = {}
                for _versionmd in _versionmds:
                    _vmd_keys[_versionmd.key] = _versionmd.value

                # # Version Metadata
                # K: versionid, key  | V: id, value
                _vmdadd = []
                for _vk, _vv in o_item.get("versions", {}).get("versionmd",{}).items():
                    _vmdref = _vmd_keys.get(_vk)
                    _vv = str(_vv)

                    if _vmdref and _vmdref != _vv:
                        # check for changes...
                        _db_sess.query(VersionMD) \
                                .filter_by(versionid=_version.id, key=_vk) \
                                .update({
                                    "value": _vv,
                                    "modified": datetime.now()
                                })

                        _stats["updates"] += 1
                    else:
                        # added new metadata
                        if not _vv:
                            continue

                        _vmdobj = VersionMD(
                            versionid = _vid,
                            key = _vk,
                            value = _vv,
                            modified = datetime.now()
                        )

                        _vmdadd.append(_vmdobj)

                _db_sess.add_all(_vmdadd)
                _db_sess.flush()

            except IntegrityError as ex:
                _db_sess.rollback()
                logger.error(ex)

            except Empty:
                break

        _db_sess.commit()
        logger.info(_stats)

    def fetch_tree(self, uri:str, zipuri:str = None, modified: datetime = None) -> dict:
        """Will return the
            uri,
            urimd,
            version,
            versionmd
        dataset."""
        _db_sess = self._db.session()

        # get the uri
        # handle zip files as well
        if zipuri:
            _uri = _db_sess.query(URIs).join(URIMD).filter(
                    URIs.uri == uri,
                    URIMD.key == 'zipfile',
                    URIMD.value == zipuri).one_or_none()
        else:
            _uri = _db_sess.query(URIs).filter(URIs.uri == uri).one_or_none()

        if not _uri:
            logger.info("ADD: %s", _uri)
        logger.info(_uri)

        _uriid = _uri.id

        # get the uri metadata
        _urimds = _db_sess.query(URIMD).filter(URIMD.uriid == _uriid).all()
        _urimd_keys = {}
        for _urimd in _urimds:
            _urimd_keys[_urimd.key] = _urimd.value

        logger.info(_urimd_keys)

        # versions
        if modified:
            logger.info("Date: %s", modified)
            _version = _db_sess.query(Versions).filter(
                    Versions.modified == modified, #o_item.get("versions", {}).get('modified'),
                    Versions.uriid == _uri.id).one_or_none()
        else:
            logger.info("User latest...")
            _subquery = _db_sess.query(
                Versions,
                func.rank().over(
                    order_by=Versions.modified.desc(),
                    partition_by=Versions.uriid
                ).label('rank')
            ).filter(Versions.uriid == _uri.id
            ).order_by(Versions.uriid).subquery()

            _version = _db_sess.query(_subquery).filter(
                    _subquery.c.rank == 1).one_or_none()

            logger.info(_version)

        logger.info(_version)
        if not _version:
            logger.info("ADD: %s", _version)

        # version metadata
        _versionmds = _db_sess.query(VersionMD).filter(VersionMD.versionid == _version.id).all()
        _vmd_keys = {}
        for _versionmd in _versionmds:
            _vmd_keys[_versionmd.key] = _versionmd.value

        logger.info(_vmd_keys)

    def catalog_artifacts(self) -> dict:
        """Will scan the sources and load a dictionary with the found files,
        it'll use the template list for extensions to use.
        <<for web addresses, it'll need a scraper built>>"""
        _data = {
            "start": time.time(),
            "processors": [],
        }


        # run through the processor list and build a queue for each one...
        _in_queues = {}
        _out_queue = mp.Queue()
        _processors = []

        for uritype, sourceobj in self._sources.items():
            logger.debug("Uri Type: %s\tObj: %s", uritype, sourceobj)

            # <TODO: add a check to match type to a processor>
            for _pname in sourceobj.get("processors"):
                if _pname not in _processors:
                    logger.debug("Name: %s\tInitiate queue", _pname)
                    _processors.append(_pname)
                    # add as a queue
                    _in_queues[_pname] = mp.Queue()

                for _src, _uri_obj in sourceobj.get("uris").items():
                    logger.debug("Name: %s\tAdd: %s", _pname, _uri_obj)
                    _in_queues[_pname].put(_uri_obj)

        # create a thread for each processor...
        for _pname, _que in _in_queues.items():
            #
            _processor = self._processors.get(_pname, {})
            logger.debug("Name: %s\n\tProcessor: %s", _pname, _processor)

            obj = _processor.get("obj")
            if not obj:
                #
                raise ClassFailedToLoad(label=_pname, message="Class didn't load!")

            obj.processorid = _processor.get("id")
            obj.in_queue(in_queue=_que)
            obj.out_queue(out_queue=_out_queue)

            # scan returns on completion
            obj.scan()

        _data["load"] = time.time(),

        # process the output
        _out_queue.put("DONE")

        self.save_queue(_out_queue)

        _data["end"] = time.time(),
        _data["processors"] = _processor

        return _data

    def parse_artifacts(self):
        """Will load the """

        logger.debug("Processors")
        for _pk, _pv in self._processors.items():
            logger.debug("Proc: %s\n\t\tVal: %s", _pk, _pv)

        logger.debug("Handlers")
        for _pk, _pv in self._urihandlers.items():
            logger.debug("Proc: %s\n\t\tVal: %s", _pk, _pv)

        _pnames = self._urihandlers.get("file")
        logger.debug(_pnames)
        if not _pnames:
            raise NoHandlersFoundForType(label="file", message="No handler found for type")

        _out_queue = mp.Queue()
        _in_queue = {}

        for _pname in _pnames:
            _proc = self._processors.get(_pname,{})
            logger.debug("Proc: %s", _proc)
            obj = _proc.get("obj")
            if not obj:
                logger.debug("missing object: %s", _pname)
                continue

            _settings = obj.avail_config()

            #<TODO: change the query set up to a dict of wueries include field lists
            #Filters
            #   <function>
            #       <tables>
            #                table alias join criteria
            #        fields  table
            #                    field: alias
            #        Pivot   table
            #                    field alias
            #        filter  table
            #                    field value | [values]
            _queries = self.get_sourcequery(filters=_settings.get("source",[]),
                        urimd=_settings.get("uri_metadata",[]),
                        versionmd=_settings.get("version_metadata",[]))

            for _qry in _queries:
                logger.debug("SQL: %s", _qry)
                _in_queue[_pname] = mp.Queue()
                _rows = self._db.sql(sql=_qry)
                for _row in _rows:
                    _in_queue[_pname].put(_row)

            obj.processorid = _proc.get("id")
            obj.in_queue(in_queue=_in_queue.get(_pname))
            obj.out_queue(out_queue=_out_queue)

            # set the helper functions...
            if _anonfunc := obj.avail_functions().get("anonimizer"):
                try:
                    _anonfunc(anonimizer=self._anonobj)
                except Exception as ex:
                    logger.warning("No Anonimizer set\n\t%s", ex)


            logger.debug("Setting tags...")
            if _tags := _proc.get("tags",{}):
                logger.debug("\tChecking proc settings...")
                for _tn in ["alias", "groups", "metadata"]:
                    if _func := obj.avail_functions().get(f"tags_{_tn}"):
                        logger.debug("\t\t%s", _tn)
                        _func(tags=_tags.get(_tn))
            elif self._tags:
                logger.debug("\tChecking generic settings...")
                for _tn in ["alias", "groups", "metadata"]:
                    if _fn := obj.avail_functions().get(f"tags_{_tn}"):
                        logger.debug("\t\t%s => %s", _tn, _fn)
                        _fn(tags=self._tags.get(f"{_tn}",[]))

            else:
                logger.debug("\tChecking default...")
                if (_fn_def := obj.avail_functions().get(f"default_alias")) and \
                        (_fn := obj.avail_functions().get(f"tags_alias")):
                    logger.debug("\t\t%s => %s", _fn_def, _fn)
                    _fn(tags=_fn_def)
                if (_fn_def := obj.avail_functions().get(f"default_groups")) and \
                        (_fn := obj.avail_functions().get(f"tags_groups")):
                    logger.debug("\t\t%s => %s", _fn_def, _fn)
                    _fn(tags=_fn_def)
                if (_fn_def := obj.avail_functions().get(f"default_metadata")) and \
                        (_fn := obj.avail_functions().get(f"tags_metadata")):
                    logger.debug("\t\t%s => %s", _fn_def, _fn)
                    _fn(tags=_fn_def)

            # initi the parser...
            logger.debug("run the parser function...")
            obj.avail_functions().get("parser")()

        # process the output
        _out_queue.put("DONE")
        self.save_queue(_out_queue)



"""A mnodule for the Rule tool code and class"""
import logging
import re

logger = logging.getLogger(__name__)

class RuleEngine():
    """The type of the engine to use to process the phrase"""
    CONTAINS = 1    # the phrase contains the expr
    EQUALS = 2      # the phrase equals the expr
    REGEX = 3       # the regex matches

class RuleState():
    """Constant class to help set the state response"""
    SINGLE = 0      # one shot
    MULTILINE = 1   # read the nect x lines as the value
    UNTIL = 2       # reas the next lines until next rule

class Rule():
    """simple class to wrap the rule
    rules: {
            "idx": <rule id>
            "name": <name for the rule>,
            "engine": "contains" | "equals" | "regex",
            "expr": <expr>,                     #
            "tags": [
                {
                    "key": <key>,               # name of the field to return
                    "value": <value>            # value to use if true
                    "regex": <regex group>      # if regex the regex group to get value from
                                                # use either value or regex
                }
            ],
            "stop": "yes" | "no" # continue or not if true
        },
    """
    def __init__(self, name:str,
            idx: int = None,
            engine: RuleEngine = RuleEngine.CONTAINS,
            expr: object = None,
            tags: list = None,
            stop: bool = True,
            state: RuleState = RuleState.SINGLE,
            options: dict = None
        ):
        """Creates the rule """
        self.name = name
        self.idx = idx
        self.engine = engine
        if self.engine == RuleEngine.REGEX:
            if isinstance(expr, str):
                self.expr = re.compile(expr)
            elif isinstance(expr, re.Pattern):
                self.expr = expr
            else:
                # bad state...
                # raise error
                pass
        else:
            self.expr = expr

        self.stop = stop
        self.tags = tags if tags is not None else {}
        self.state = state
        self.options = options

        # save out defaults
        self.ignorecase = self.options.get("ignorecase", True) if self.options is not None else True

    def run(self, phrase: str) -> dict:
        """Processes the rule against the parsed string
        return dict
            passed: True | False
            results: [key: value]
            stop: True | False"""
        data = {
            "passed": False,
            "results": {},
            "stop": self.stop,
        }
        if phrase is None or len(phrase) == 0:
            return data

        # check what type of rule to use
        if self.engine == RuleEngine.REGEX:
            if self.options.get("findall", False) is True:
                # use the find all to recover alll matchin
                # <TODO: add finall code here>
                if m := re.findall(self.expr, phrase):
                    for tagvalue in self.tags:
                        tagkey = tagvalue.get("key")
                        if grpkey := tagvalue.get("group"):
                            # use the regex group
                            resvalue = m.group(grpkey)
                        else:
                            resvalue = tagvalue.get("value")
                        if tagkey not in data.get("results", {}):
                            data["results"][tagkey] = set()
                        data["results"][tagkey].add(resvalue)

            elif m := re.match(self.expr, phrase):
                data["passed"] = True
                # populate the values if any
                for tagvalue in self.tags:
                    tagkey = tagvalue.get("key")
                    if grpkey := tagvalue.get("group"):
                        # use the regex group
                        resvalue = m.group(grpkey)
                    else:
                        resvalue = tagvalue.get("value")
                    data["results"][tagkey] = resvalue

        elif self.engine == RuleEngine.CONTAINS:
            if self.ignorecase is True:
                if self.expr.lower() in phrase.lower():
                    data["passed"] = True
            else:
                if self.expr == phrase:
                    data["passed"] = True
            # populate the values if any
            if data.get("passed") is True:
                for tagvalue in self.tags:
                    tagkey = tagvalue.get("key")
                    data["results"][tagkey] = tagvalue.get("value")

        elif self.engine == RuleEngine.EQUALS:
            #
            if self.ignorecase is True:
                if self.expr.lower() in phrase.lower():
                    data["passed"] = True
            else:
                if self.expr == phrase:
                    data["passed"] = True
            # populate the values if any
            if data.get("passed") is True:
                for tagvalue in self.tags:
                    tagkey = tagvalue.get("key")
                    data["results"][tagkey] = tagvalue.get("value")

        # <TODO: add the conditional and state rules here...

        return data

class RulesTool:
    """ manages and runs the rules against a
    string
    """

    # instatitate as a class variable
    name = "names"
    version = "0.0.1"

    def __init__(self) -> None:
        self.rules = []

    def add_rule(self, rule: Rule):
        """Adds a regex to the class"""
        self.rules.append(rule)

    def run(self, phrases: list):
        """process the rules against each line and
        return the result"""
        results = []
        for phrase in phrases:
            for rule in self.rules:
                result = rule.run(phrase=phrase)
                passed = result.get("passed", False)
                stop = result.get("stop", False)

                if passed is True:
                    results.append({
                        "rule": rule,
                        "phrase": phrase,
                        "result": {
                            "passed": passed,
                            "stop": stop,
                            "results": result.get("results")
                        }
                    })
                if stop:
                    continue

        return results



"""This is a class module for the anonimizer"""
import logging
import random
import string
import uuid
from datetime import date, datetime

logger = logging.getLogger(__name__)

# create unique id
class TagAnon():
    """A class to handle the anonimizing and past tags and values."""
    def __init__(self, tags: dict) -> None:
        """Code with 4 base functions:
            name - returns a ([a-z0-9]{3})-([A-Z0-9]{9}) string
            UUID - returns a GUID
            Date - Will return the <YYYY>-01-01
            Seq  - Will return an incremental number
        """
        self.func_map = {
            "name": self._get_uname,
            "uuid": self._get_uuid,
            "date": self._get_udate,
            "seq": self._get_useq,
        }

        self.tag_map = {}
        self.tag_lookup = {
            "__names": []
        }

        for k_tag, v_tag in tags.items():
            func = self.func_map.get(v_tag,"seq")

            # initialize the function map
            if k_tag not in self.tag_map:
                self.tag_map[k_tag] = func

            # initiialize the lookups
            if k_tag not in self.tag_lookup:
                self.tag_lookup[k_tag] = {}

    def load_cache(self, tag_lookup: dict) -> None:
        """Allows for previsously remembered lookup to be loaded,
        *********************************************************
        this cache will potentially have PII data in it and
        should only be used for a batch of imports
        *********************************************************
        """
        self.tag_lookup = tag_lookup

    def fetch_cache(self) -> dict:
        """
        return the dict of the lookups
        *********************************************************
        this cache will potentially have PII data in it and
        should only be used for a batch of imports
        *********************************************************
        The cache is a pickle file of a dict
        """
        return self.tag_lookup

    def is_pii(self, tag: str) -> bool:
        """Will check if the tag is on the PI (Anon) list"""
        return tag in self.tag_lookup

    def get_anon(self, tag: str, value: str) -> str:
        """For the past tag and value, it will add the value to a lookup
        and return the new value.
        If the same tag|value combo is past in, it'll return the prior value
        for the lifetime of the class instance.
        The cache can be saved and loaded"""
        #self._fix_missing(tag=tag)
        if tag not in self.tag_lookup:
            return value

        new_value = self.tag_lookup.get(tag, {}).get(value)
        if not new_value:
            new_value = self.tag_map.get(tag, self._get_useq)(tag=tag, value=value)
            self.tag_lookup[tag][value] = new_value

        return new_value

    def add_tag(self, tag: str, func: str) -> None:
        """Will add a tag and the associated function to the lookup tables"""
        #  handle the missing value cases
        if tag not in self.tag_map:
            self.tag_map[tag] = self.func_map.get(func, "def")

        if tag not in self.tag_lookup:
            self.tag_lookup[tag] = {}

    def _get_uuid(self, tag: str, value:str) -> str:
        """Returns a UUID"""
        return str(uuid.uuid4())

    def _get_uname(self, tag: str, value: str) -> str:
        """Build a random string of 3-9"""
        new_value = self.get_random_string(prefix_len=3, body_len=9)
        while True:
            if new_value in self.tag_lookup.get("__names"):
                new_value = self.get_random_string(prefix_len=3, body_len=9)
                continue

            self.tag_lookup["__names"].append(new_value)
            break

        return new_value

    def _get_udate(self, tag: str, value: str) -> date:
        """Zeros the month and day to 01-01"""
        # default to 1900-01-01
        new_dob = date(1900,1,1)
        try:
            dob = datetime.strptime(value, "%Y%m%d")
            new_dob = date(dob.year, 1, 1)

        except ValueError:
            logger.error("bad date, %s -> %s, using default", tag, value)

        return new_dob.strftime("%Y%m%d")

    def _get_useq(self, tag: str, value: str) -> date:
        """Return a simple incremental ID"""
        # index as seen
        return len(self.tag_lookup.get(tag, {})) + 1
        pass

    def get_random_string(self, prefix_len: int, body_len: int) -> str:
        """Function to build a random string [prefex]-[body]"""
        # choose from all lowercase letter
        letters = string.ascii_uppercase
        numb = string.hexdigits
        str_02 = ''.join(random.choice(numb + letters + numb) for i in range(body_len))

        return "{}-{}".format(''.join(random.choice(numb) for i in range(prefix_len)), str_02)


if __name__ == "__main__":
    tags_anon = {
        #"PatientID":        "seq",          # (0010,0020)
        "PatientName":      "name",
        "ReviewerName":     "name",         # (300E,0008)
        "PatientBirthDate": "date",         # (0010,0030)
    }

    test_tags = {
        "PatientID": [999, 888, 777, 666, 555, 444],
        "PatientName": ['alice', 'bob', 'chris', 'david', 'eric', 'fred'],
        "ReviewerName": ['zeebeedee', 'yazoo','xavier','william',' una'],
        "PatientBirthDate": ['19050512', '20221231', '20000101', '14000505']
    }

    anon_obj = TagAnon(tags=tags_anon)

    for idx in range(1, 50):
        tag = random.choice(list(tags_anon.keys()))
        value = random.choice(test_tags.get(tag,[]))
        res = anon_obj.get_anon(tag=tag, value=value)
        print("{:20}|\t{:>15}\t{:>15}".format(tag, value, res))

    anon_obj = None



""" A module for fucntions and classes to help manage run time
    loaded modules.
    Returns a set of dictionaries with values as the class or functions """

import importlib
import inspect
import logging
import os
import sys

logger = logging.getLogger(__name__)

class ModuleFailedToLoad(Exception):
    """A simple exception to raise module load error"""
    def __init__(self, label: str, message: str) -> None:
        super.__init__()
        self.label = label
        self.message = message

def load_modules(folderpath: str):
    _data = {}
    for _folder, _, _files in os.walk(os.path.abspath(folderpath)):
        for _file in _files:
            if not _file.lower().endswith(".py"):
                continue
            if _file == "__init__.py":
                continue

            _mod_data = load_modulefile(_folder, _file)
            for _fk, _fv in _mod_data.items():
                _data[_fk] = _fv
    return _data

def load_modulefile(folder: str, file: str, funcs: list = None):
    """Loads a module from a relative or abs path"""
    _data = {}
    _pfile = os.path.abspath(os.path.join(folder, file))
    _m_name = file.replace(".py", "")
    _spec = importlib.util.spec_from_file_location(_m_name, _pfile)
    logger.debug("Spec:   %s", _spec)

    _module = importlib.util.module_from_spec(_spec)
    _spec.loader.exec_module(_module)
    logger.debug("Module: %s", _module)
    logger.debug("Dir:    %s", dir(_module))
    _members = dict(inspect.getmembers(_module))

    if funcs:
        for _fn in funcs:
            logger.debug("Func: %s", _fn)
            _data[_fn] = _members.get(_fn)
    else:
        logger.debug("Inspect: %s", _members)
        sys.modules[_spec.name] = _module
        for _fn, _fv in _members.items():
            if _fn.startswith("__"):
                continue

            _data[_fn] = _fv

    return _data

def load_module(module: str, funcs: list = None) -> dict:
    """Loads a module based on the module path rather than
    file path, if a func list is provided ti will only
    return the object pointers
    """
    _data = {}
    try:
        components = module.split('.')
        logger.debug("Path: %s Funcs: %s", components, funcs)

        if funcs:
            # assum full path
            mod = __import__(".".join(components), fromlist=funcs)
            logger.debug("Module: %s", dir(mod))

            for _fn in funcs:
                _data[_fn] = getattr(mod, _fn)
                logger.debug(_data[_fn])

        else:
            # assume the last element is the function or class
            mod = __import__(".".join(components[:-1]), fromlist=components[-1])
            _data[components[-1]] = getattr(mod, components[-1])

            # finds the classes in the module..
            # check each class and add to processor list...
            logger.debug(_data[components[-1]])

    except ImportError as ex:
        raise ModuleFailedToLoad(label=module,
                message="module could not be loaded") from ex

    return _data



"""this will scan a file system and return the file details"""
import logging
import multiprocessing as mp
import os
import re
import threading as td

from queue import Empty
from socket import gethostname

from lost_cat.utils.path_utils import SourceDoesNotExist, get_filename, get_file_metadata
from lost_cat.processors.base_processor import BaseProcessor
from lost_cat.utils.path_utils import scan_files, SourceNotHandled, get_machine

logger = logging.getLogger(__name__)

class FileScanner(BaseProcessor):
    """This class will perform serveral operations:
            process an acceptable uri
            scan a uri and catalog found items there
            fetch an item from a uri
            upload an item to a uri
    """

    def __init__(self, settings: dict = None):
        """"""
        super().__init__(settings=settings)
        self._version = "0.0.2"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"
        self._semiphore = f"DONE: {self._name}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = FileScanner.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Semiphore: %s", self._semiphore)
        logger.debug("Settings: %s", self.settings)

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "build": self.build_path,
            "scanner": self.scan,
            #"fetch": self.fetch,
            #"upload": self.upload,
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{
                "splitfolders": True,
                "splitextension": True,
                "stats": True,
            },
            "uritypes": [
                "folder",
            ],
            "filters":[
                {
                    "table": "URIs",
                    "field": "root",
                    "select": True
                }
            ],
            "threads": {
                "count": 1,
                "stop": True, # tells the system to stop when the in queue is empty
                "timeout": 2
            }
        }

    def build_path(self, uri: str, hint: str = "folder") -> dict:
        """will accept a uri and will turn it to a dictionary
        for later use.
        and return a dictionary
            "type": str
            "uri": uri

            "root": str
            "folders": list
            "filename": str
            "ext": str
        """

        src = {
            "uri": uri,
            "type": None
        }

        logger.debug("Check: %s\t:%s\t:%s ", r"[\\\/\.]+", uri, re.search(r"[\\\/\.]+", uri))
        try:
            _fmd = get_file_metadata(uri=uri, options={"stats": True})
            logger.debug("File Stats %s\n\t\t%s", uri, _fmd)

        except SourceDoesNotExist:
            logger.debug("Provide uri does not exist %s", uri)
            return None

        if re.search(r"[\\\/]+", uri):
            _uri = os.path.expandvars(os.path.expanduser(uri))
            _drv, _path = os.path.splitdrive(uri)

            if os.path.exists(_uri):
                if os.path.isdir(_uri):
                    src["type"] = "folder"

                elif os.path.isfile(_uri):
                    src["type"] = "file"
            else:
                src["type"] = hint

            if src["type"] == "file":
                _path, _filename = os.path.split(_path)
                _name, _ext = os.path.splitext(_filename)
                src["name"] = _name
                src["ext"] = _ext.lower()

            _folders = []
            while len(_path) > 1:
                _path, _fld = os.path.split(_path)
                if len(_fld) > 0:
                    _folders.append(_fld)

            _folders.reverse()
            src["folders"] = _folders
            src["root"] = f"{_drv}{os.sep}"
            src["domain"] = _fmd.get("domain")

        else:
            raise SourceNotHandled(uri=uri,
                    message="provided uri could not be understood by parser!")

        return src

    def scan(self):
        """scan the file system use multprocess to
        scan each item in the queue"""
        use_threads = self.settings.get("threads",{}).get("count",5)

        for t_idx in range(use_threads):
            logger.info("Thread: %s",t_idx)
            scan_q = td.Thread(target=self.scan_fs)
            scan_q.start()
            scan_q.join()

    def scan_fs(self):
        """A file scanner function, quick and dirty"""
        t_settings = self.settings.get("threads",{})
        _hostname = gethostname()

        while self.input:
            try:
                q_item = self.input.get(timeout=t_settings.get("timeout")) if self.input else None
                if q_item == self.semiphore:
                    break

                # if the user wants to kill the queue if there are not entries...
                # requires timeout set, otherwise the queue get blocks...
                if not q_item and self.settings.get("threads",{}).get("stop"):
                    break

                uri = q_item
                if isinstance(q_item, dict):
                    uri = q_item.get("uri")

                domainmd = get_machine()

                for fs_obj in scan_files(uri=uri, options=self.settings.get("options",{})):
                    # convert to the
                    # URIs
                    #   -> Metadata
                    #   -> Versions
                    #       -> metadata
                    # {
                    #   'type': | 'root': | 'folder': | 'name': | 'ext': |
                    #   'folders': | 'accessed': | 'modified': | 'created': | 'size': | 'mode':
                                            # add domain md to first record

                    _uri = get_filename(fs_obj)
                    _data = {
                        "domain": _hostname,
                        "processorid": self.processorid,
                        "uriid_source": q_item.get("uriid"),
                        "uri_type": fs_obj.get("type"),
                        "uri": _uri, #fs_obj.get("uri"),
                        "metadata": {
                            "ext": fs_obj.get("ext"),
                            "created": fs_obj.get("created"),
                            "name": fs_obj.get("name"),
                        },
                        "versions": {
                            "modified": fs_obj.get("modified"),
                            "size": fs_obj.get("size"),
                            "checksum": fs_obj.get("hash", {}).get("SHA1"),
                        }
                    }
                    if domainmd:
                        _data["domainmd"] = domainmd
                        domainmd = None

                    self.output.put(_data)

                    for zs_obj in fs_obj.get("files", []):
                        # dict :    zipfile
                        #           path
                        #           folder
                        #           name
                        #           size
                        #           ext

                        logger.debug(zs_obj)
                        _data = {
                            "domain": _hostname,
                            "processorid": self.processorid,
                            "uriid_source": q_item.get("uriid"),
                            "uri_type": "zipfile",
                            "uri": zs_obj.get("path"),
                            "metadata": {
                                "ext": zs_obj.get("ext"),
                                # <TODO: get the create / modified dates from the zipfile itself>
                                "created": fs_obj.get("created"),
                                "name": zs_obj.get("name"),
                                "zipfile": _uri,
                            },
                            "versions": {
                                # <TODO: get the create / modified dates from the zipfile itself>
                                "modified": fs_obj.get("modified"),
                                "size": zs_obj.get("size")
                            }
                        }
                        self.output.put(_data)

            except Empty:
                break



"""A module to wrap the database functions and allow the syst4em to wru"""
import logging

from lost_cat.database.schema import Base
from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.orm import sessionmaker, session as sqlsession

logger = logging.getLogger(__name__)

class DBEngine():
    """"""
    def __init__(self, connString: str = "sqlite:///lost-cat.db") -> None:
        """"""
        self.engine = create_engine(connString, echo=False)

        # create the mapping objects...
        Base.metadata.create_all(self.engine)
        self.metadata = MetaData()
        #self.session = sessionmaker(bind=self.engine)()

    def session(self) -> sqlsession:
        """"""
        _session = sessionmaker(bind=self.engine)
        return _session()

    def close(self):
        """"""
        pass

    def table(self, tablename: str):
        """"""
        return Table(tablename, self.metadata, autoload=True, autoload_with=self.engine)

    def tables_list(self) -> dict:
        """"""
        conn = self.engine.connect()
        _data = []
        _res = conn.execute("SELECT name FROM sqlite_master WHERE type='table';")
        _keys = _res.keys()
        for _row in list(_res):
            _rout = {}
            for _rid, _rc in enumerate(_keys):
                _rout[_rc] = _row[_rid]
            _data.append(_rout)

        return _data

    def sql(self, sql: str) -> list:
        """Returns a dictionary of objects"""
        _data = []
        _db_sess = self.session()
        _res = _db_sess.execute(sql)
        _keys = _res.keys()
        logger.debug("COLUMNS: %s", _keys)
        for _row in list(_res.fetchall()):
            _rout = {}
            for _rid, _rc in enumerate(_keys):
                _rout[_rc] = _row[_rid]
            _data.append(_rout)
            #logger.debug("ROW: %s", _rout)

        return _data



"""This module provides a set of functions for handle paths,
file scanning and zip file handling.
@author: Dreffed
copyright adscens.io 2022 / thoughtswin systems 2022
"""
import hashlib
import io
import logging
import os
import platform
import re
import tarfile
import time
import uuid
import zipfile
from datetime import datetime
from time import mktime

import psutil

try:
    from os import DirEntry, scandir
except ImportError:
    from scandir import scandir, DirEntry

from socket import gethostbyname, gethostname
from urllib.parse import urlparse
from validators import url as val_url

from lost_cat.utils.phrase_utils import PhraseTool

logger = logging.getLogger(__name__)

class SourceDoesNotExist(Exception):
    """A class to raise the missing file"""

class SourceNotValid(Exception):
    """A simple exception class to catch edge casses"""
    def __init__(self, uri:str, message:str, *args: object) -> None:
        self.uri = uri
        self.message = message
        super().__init__(*args)

class SourceNotHandled(Exception):
    """A simple exception class to catch edge casses"""
    def __init__(self, uri:str, message:str, *args: object) -> None:
        self.uri = uri
        self.message = message
        super().__init__(*args)

def func_switch_zip(ext: str, op_label: str) -> object:
    """A sweithc dict to enable the selection of the
    function for the zip file handlers, hard coded :("""
    func = {
        ".zip": {
            "open": open_zip,
            "scan": scan_zip,
            "fetch": fetch_zip
        },
        ".tar.gz": {
            "open": open_tar,
            "scan": scan_tar,
            "fetch": fetch_tar
        }
    }
    return func.get(ext,{}).get(op_label)

def get_machine() -> dict:
    """Will return a dictionary of the machine details"""
    try:
        _data={}
        _data['platform']=platform.system()
        _data['platform-release']=platform.release()
        _data['platform-version']=platform.version()

        _data['architecture']=platform.machine()
        _data['hostname']=gethostname()
        _data['ip-address']=gethostbyname(gethostname())

        _data['mac-address']=':'.join(re.findall('..', '%012x' % uuid.getnode()))
        _data['processor']=platform.processor()
        _data['ram']=str(round(psutil.virtual_memory().total / (1024.0 **3)))+" GB"
    except Exception as e:
        logging.exception(e)

    return _data

def split_folder(path: str) -> list:
    """Will split the folder path in to a list of folders"""
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == os.sep:
            allparts.insert(0, parts[1])
            break
        elif parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def build_path(uri: str, path_type: str = "file") -> dict:
    """Will take a path, and split into components
    and return a dictionary
    {
        "root": str
        "type": str
        "folders": list
        "name": str
        "ext": str
    }"""

    src = {
        "root": None,
        "type": None,
        "folders": []
    }

    logger.debug("Check: %s\t:%s\t:%s ", r"[\\\/\.]+", uri, re.search(r"[\\\/\.]+", uri))

    if val_url(uri):
        src["type"] = "http"

        url_obj = urlparse(uri)
        src["scheme"] = url_obj.scheme
        src["netloc"] = url_obj.netloc
        src["path"] = url_obj.path
        src["params"] = url_obj.params
        src["query"] = url_obj.query
        src["fragment"] = url_obj.fragment

    elif re.search(r"[\\\/\.]+", uri):
        uri = os.path.expandvars(os.path.expanduser(uri))
        drv, path = os.path.splitdrive(uri)

        if os.path.exists(uri):
            if os.path.isdir(uri):
                src["type"] = "folder"

            elif os.path.isfile(uri):
                src["type"] = "file"
        else:
            src["type"] = path_type

        if src["type"] == "file":
            path, filename = os.path.split(path)
            name, ext = os.path.splitext(filename)
            src["name"] = name
            src["ext"] = ext.lower()

        folders = []
        while len(path) > 1:
            path, fld = os.path.split(path)
            if len(fld) > 0:
                folders.append(fld)

        folders.reverse()
        src["folders"] = folders
        src["root"] = f"{drv}{os.sep}"

    else:
        raise SourceNotHandled(uri=uri, message="provided uri could not be understood by parser!")

    return src

def get_filename(file_dict: dict) -> str:
    """Will accept a json object of file details and return the
    full filename
        {
            "root":     str
            "folder":   str     (optional)
            "folders":  []      (optional)
            "name":     str     (optional)
            "ext":              (optional) expects "." prefix
        }
        <TODO: Add handler for type = http etc.>
    """

    #logger.debug("F: %s", file_dict)
    if file_dict.get("root") == ".":
        file_dict["root"] = os.getcwd()
        logger.debug("using current drive %s", file_dict.get("root"))

    if "folders" in file_dict:
        _path = os.path.join(file_dict.get("root"), *file_dict.get("folders",[]))
    else:
        _path = os.path.join(file_dict.get("root"), file_dict.get("folder",""))

    # expand the path for user and env vars
    _path = os.path.expandvars(os.path.expanduser(_path))

    if "name" in file_dict:
        _path = os.path.join(_path, "{}{}".format(file_dict.get("name"), file_dict.get("ext","")))

    return _path

def get_file_metadata(uri: str, options: dict = None) -> dict:
    """return dict of file meats data based on the options passed
        "name": the file name of the file, if "splitextension" specified it is
                just the filename not extension
        "ext": the dotted extension of the file
        "folder"L the folder path of the file

        optional output:
            "folders": an array of the folder names, option "split"
            date and size details: option "stats"
                "modified"
                "accessed"
                "size"
                "bytes": szie in mb, gb, etc.

    """
    if os.path.exists(uri):
        if os.path.isfile(uri):
            _f_type = "file"
        elif os.path.isdir(uri):
            _f_type = "folder"
        else:
            _f_type = "unclassified"
    else:
        _f_type = "unclassified"
        
    _drv, _path = os.path.splitdrive(uri)

    if os.path.isfile(uri):
        _path = os.path.dirname(_path)
        _filename = os.path.basename(uri)
        _fname, _ext = os.path.splitext(_filename)
        _file_dict =  {
                "domain":   gethostname(),
                "type":     _f_type,
                "root":     f"{_drv}{os.sep}",
                "folder" :  _path,
                "name" :    _filename,
                "ext":      _ext.lower()
        }
    else:
        _file_dict =  {
                "domain":   gethostname(),
                "type":     _f_type,
                "root":     f"{_drv}{os.sep}",
                "folder" :  _path
        }

    if options and options.get("splitextension"):
        _file_dict["name"] = _fname

    if options and options.get("splitfolders"):
        _file_dict["folders"] = split_folder(_path)

    if options and options.get("stats"):
        file_stats = os.stat(uri)
        _file_dict["accessed"] = datetime.fromtimestamp(mktime(time.localtime(file_stats.st_atime)))
        _file_dict["modified"] = datetime.fromtimestamp(mktime(time.localtime(file_stats.st_mtime)))
        _file_dict["created"] = datetime.fromtimestamp(mktime(time.localtime(file_stats.st_ctime)))
        _file_dict["size"] = file_stats.st_size
        _file_dict["mode"] = file_stats.st_mode
        _file_dict["device"] = file_stats.st_dev

    return _file_dict

def make_hash(uri: str, buff_size: int = 65536) -> dict:
    """ Will hash the files for both MD5 and SHA1 and return a dict of the hashes"""
    hashes = {}
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()

    try:
        if os.path.exists(uri):
            #logger.debug('%s', uri)
            with open(uri, 'rb') as f_io:
                while True:
                    d_bytes = f_io.read(buff_size)
                    if not d_bytes:
                        break
                    md5.update(d_bytes)
                    sha1.update(d_bytes)

            hashes['MD5'] = md5.hexdigest()
            hashes['SHA1'] = sha1.hexdigest()
        else:
            hashes['MD5'] = 'Missing file'
            hashes['SHA1'] = 'Missing file'

    except OSError as ex:
        logger.error('ERROR: [%s]\n%s', uri, ex)

        hashes['MD5'] = 'ERROR'
        hashes['SHA1'] = 'ERROR'

    return hashes

def fast_scan(uri: str) -> DirEntry:
    """A fast scan routine to generate a
    simple file objecvt set """
    for os_obj in scandir(path=uri):
        try:
            if os_obj.is_dir(follow_symlinks=False):
                yield from fast_scan(uri=os_obj.path)
            else:
                if os_obj.is_file():
                    #logger.debug(os_obj)
                    yield os_obj

        except PermissionError:
            logger.error("Permission Error: %s", os_obj.path)

def scan_files(uri: str, options: dict = None) -> dict:
    """Will scan the folder and walk the files and folders below
    yields the found file"""
    for os_obj in fast_scan(uri=uri):
        try:
            drv, path = os.path.splitdrive(os_obj.path)
            dirpath = os.path.dirname(path)
            filename = os.path.basename(path)
            fname, ext = os.path.splitext(filename)

            f_type = "file" if os_obj.is_file() else "folder"

            file_dict =  {
                    "type":f_type,
                    "uri": os_obj.path,
                    "root": f"{drv}{os.sep}",
                    "folder" : dirpath,
                    "name" : filename,
                    "ext": ext.lower()
            }

            if options and options.get("splitextension"):
                file_dict["name"] = fname

            if options and options.get("splitfolders"):
                file_dict["folders"] = split_folder(dirpath) #dirpath.split(os.sep) #

            # filter the files...
            if not is_include(file_dict=file_dict, options=options):
                continue

            if options and options.get("stats"):
                file_stats = os.stat(uri)
                file_dict["accessed"] = datetime.fromtimestamp(
                        mktime(time.localtime(file_stats.st_atime))) #
                file_dict["modified"] = datetime.fromtimestamp(
                        mktime(time.localtime(file_stats.st_mtime)))
                file_dict["created"] = datetime.fromtimestamp(
                        mktime(time.localtime(file_stats.st_ctime)))
                file_dict["size"] = file_stats.st_size
                file_dict["mode"] = file_stats.st_mode

            if options and options.get("profile"):
                p_obj = PhraseTool(in_phrase=file_dict.get("name",""))
                file_dict["profile"] = p_obj.get_metadata()

            # handel the options for hash
            if options and options.get("generatehash"):
                maxsize = options.get("maxhashsize", 0)
                if maxsize == 0 or maxsize >= file_dict.get("size",0):
                    file_dict["hash"] = make_hash(uri=os_obj.path)

            op_func = func_switch_zip(ext, "scan")
            if op_func:
                zfiles = op_func(uri=os_obj.path)
                file_dict["files"] = zfiles
                yield file_dict

            else:
                yield file_dict

        except SourceDoesNotExist as ex:
            logger.error('URI: %s %s', os_obj.path, ex)
        except zipfile.BadZipFile as ex:
            logger.error('URI: %s %s', os_obj.path, ex)
        except PermissionError as ex:
            logger.error("Permission Error: %s %s", os_obj.path, ex)


def is_include(file_dict: dict, options: dict = None) -> bool:
    """will use the filter conditions and if all filters are matched
    will return true
    ...
        filter
            ext     : list of extensions to check against
            regex   : regex phrase
    """
    if not options:
        return True

    filter_config = options.get("filter")
    if not filter_config:
        return True

    output = {}
    if "exts" in filter_config:
        output["ext"] = 0
        if file_dict.get("ext","") in filter_config.get("exts", []):
            output["ext"] = 1

    if "regex" in filter_config:
        output["regex"] = 0
        filter_reg = re.compile(filter_config.get("regex", ""))
        m_re = filter_reg.match(file_dict.get("name",""))
        if m_re:
            output["regex"] = 1

    return len(output) == sum(output.values)

def open_zip(uri: str) -> zipfile.ZipFile:
    """Will open a zip file and return the file handle"""
    return zipfile.ZipFile(uri)

def open_tar(uri: str) -> tarfile.TarFile:
    """Will open a tar file and return the handle"""
    return tarfile.open(uri, mode="r")

def scan_zip(uri: str) -> dict:
    """Will scan the zip file and return the file details"""
    zip_file = zipfile.ZipFile(uri)
    files = []

    for szf in zip_file.infolist():
        if not szf.is_dir():
            filepath = szf.filename
            dirpath, fullname = os.path.split(filepath)
            filename, ext = os.path.splitext(fullname)
            sub_file = {
                "zipfile": uri,
                "path": filepath,
                "folder": dirpath,
                "name": filename,
                "size": szf.file_size,
                "ext": ext.lower(),
            }
            files.append(sub_file)

    return files

def scan_tar(uri: str) -> dict:
    """Will handle the targz file"""
    tar_file = tarfile.open(uri, mode="r")
    files = []
    for szf in tar_file.getmembers():
        # process only files...
        if not szf.isfile():
            continue

        filepath = szf.name
        dirpath, fullname = os.path.split(filepath)
        filename, ext = os.path.splitext(fullname)
        yr, mm, dd, hh, mn, ss = szf.date_time
        #mod_date = datetime(year=yr, month=mm, day=dd, hour=hh, minute=mn, second=ss)
        sub_file = {
            "zipfile": uri,
            "path": filepath,
            "folder": dirpath,
            "name": filename,
            "size": szf.size,
            "modified": datetime(year=yr, month=mm, day=dd, hour=hh, minute=mn, second=ss),
            "ext": ext.lower(),
        }
        files.append(sub_file)

    return files

def fetch_zip(file_obj: zipfile.ZipFile, item_path: str) -> io.BytesIO:
    """for a given zip file, return the item"""
    _data = file_obj.read(item_path)
    return io.BytesIO(_data)

def fetch_tar(file_obj: tarfile.TarFile, item_path: str) -> io.BytesIO:
    """For a given tarfile return the item"""
    _data = file_obj.read(item_path)
    return io.BytesIO(_data)
